/*
    This program implements a simple synchronization pattern
    between two threads. The synchronization pattern is to keep
    the two threads "alternating" with each other. Each thread
    does a sequence of "computations," where each computation can
    take a different amount of time, and the threads are synchronized
    so that each thread begins a computation exactly when the
    other threads ends a computation. In addition, whichever thread
    is ready to run first becomes the "first" thread in the alternating
    pattern. In other words, this program does not favor thread1 or
    thread2 as the prefered "first" thread.

    This program uses a mutex, a shared variable, and two local variables
    to implement the synchronization pattern.
*/
#include <stdio.h>
#include <windows.h>

#define RUN_TIME 30  // run time in seconds

// global variables
int runFlag = TRUE;

HANDLE mutex;        // used to protect a critical section
int globalFlag = 0;  // used to flag which thread is "first"

int main( int argc, char *argv[])
{
   // thread function prototypes
   DWORD WINAPI thread1(LPVOID);
   DWORD WINAPI thread2(LPVOID);

   // create a Mutex object (in signaled state)
   mutex = CreateMutex(NULL, FALSE, NULL); // signaled

   // create two threads
   printf("alternating1: Starting threads\n");
   CreateThread(NULL, 0, thread1, NULL, 0, NULL);
   CreateThread(NULL, 0, thread2, NULL, 0, NULL);

   Sleep(RUN_TIME*1000); // let the threads run for a while
   runFlag = FALSE;      // now stop the threads
   Sleep(200);           // give the threads time to stop

   printf("\nThreads stopped\n");

   return 0;
}//main


DWORD WINAPI thread1(LPVOID threadParam)
{
   int i;
   int workTime, delayTime;
   int localFlag;
   srand(GetTickCount());  // seed the random number generator

   delayTime = 2;
   fprintf(stderr, "Thread1 delayed for %d seconds\n", delayTime);
   Sleep(delayTime*1000);

   WaitForSingleObject(mutex, INFINITE);  // enter critical section
   localFlag = globalFlag;                // thread that gets value 0 is "first"
   globalFlag = 1;
   ReleaseMutex(mutex);

   while(runFlag)
   {
      while( globalFlag == localFlag )  // wait for turn
         Sleep(0);
      workTime = 1 + rand()%9;          // 1 to 9 seconds of work time
      fprintf(stderr, "\nThread1 working for %d seconds", workTime);
      for (i = 0; i < workTime; i++)
      {
         fprintf(stderr, ".");
         Sleep(1000);
      }
      fprintf(stderr, "done");
      globalFlag = 1 - globalFlag;  // toggle 0 to 1 and 1 to 0
   }
   return 0;
}//thread1


DWORD WINAPI thread2(LPVOID threadParam)
{
   int i;
   int workTime, delayTime;
   int localFlag;
   srand(GetTickCount());  // seed the random number generator

   delayTime = 10;
   fprintf(stderr, "Thread2 delayed for %d seconds\n", delayTime);
   Sleep(delayTime*1000);

   WaitForSingleObject(mutex, INFINITE);  // enter critical section
   localFlag = globalFlag;                // thread that gets value 0 is "first"
   globalFlag = 1;
   ReleaseMutex(mutex);

   while(runFlag)
   {
      while( globalFlag == localFlag )  // wait for turn
         Sleep(0);
      workTime = 1 + rand()%9;          // 1 to 9 seconds of work time
      fprintf(stderr, "\nThread2 working for %d seconds", workTime);
      for (i = 0; i < workTime; i++)
      {
         fprintf(stderr, ".");
         Sleep(1000);
      }
      fprintf(stderr, "done");
      globalFlag = 1 - globalFlag;  // toggle 0 to 1 and 1 to 0
   }
   return 0;
}//thread2
